"use strict";
// Copyright IBM Corp. 2017. All Rights Reserved.
// Node module: @loopback/rest
// This file is licensed under the MIT License.
// License text available at https://opensource.org/licenses/MIT
Object.defineProperty(exports, "__esModule", { value: true });
const context_1 = require("@loopback/context");
const stream = require("stream");
const debug = require('debug')('loopback:rest:router:metadata');
const REST_METHODS_KEY = 'rest:methods';
const REST_METHODS_WITH_PARAMETERS_KEY = 'rest:methods:parameters';
const REST_PARAMETERS_KEY = 'rest:parameters';
const REST_CLASS_KEY = 'rest:class';
const REST_API_SPEC_KEY = 'rest:api-spec';
/**
 * Decorate the given Controller constructor with metadata describing
 * the HTTP/REST API the Controller implements/provides.
 *
 * `@api` can be applied to controller classes. For example,
 * ```
 * @api({basePath: '/my'})
 * class MyController {
 *   // ...
 * }
 * ```
 *
 * @param spec OpenAPI specification describing the endpoints
 * handled by this controller
 *
 * @decorator
 */
function api(spec) {
    return context_1.ClassDecoratorFactory.createDecorator(REST_CLASS_KEY, spec);
}
exports.api = api;
/**
 * Build the api spec from class and method level decorations
 * @param constructor Controller class
 */
function resolveControllerSpec(constructor) {
    debug(`Retrieving OpenAPI specification for controller ${constructor.name}`);
    let spec = context_1.MetadataInspector.getClassMetadata(REST_CLASS_KEY, constructor);
    if (spec) {
        debug('  using class-level spec defined via @api()', spec);
        spec = context_1.DecoratorFactory.cloneDeep(spec);
    }
    else {
        spec = { paths: {} };
    }
    let endpoints = context_1.MetadataInspector.getAllMethodMetadata(REST_METHODS_KEY, constructor.prototype) || {};
    endpoints = context_1.DecoratorFactory.cloneDeep(endpoints);
    for (const op in endpoints) {
        debug('  processing method %s', op);
        const endpoint = endpoints[op];
        const verb = endpoint.verb;
        const path = endpoint.path;
        let endpointName = '';
        /* istanbul ignore if */
        if (debug.enabled) {
            const className = constructor.name || '<AnonymousClass>';
            const fullMethodName = `${className}.${op}`;
            endpointName = `${fullMethodName} (${verb} ${path})`;
        }
        let operationSpec = endpoint.spec;
        if (!operationSpec) {
            // The operation was defined via @operation(verb, path) with no spec
            operationSpec = {
                responses: {},
            };
            endpoint.spec = operationSpec;
        }
        debug('  operation for method %s: %j', op, endpoint);
        debug('  processing parameters for method %s', op);
        let params = context_1.MetadataInspector.getAllParameterMetadata(REST_PARAMETERS_KEY, constructor.prototype, op);
        if (params == null) {
            params = context_1.MetadataInspector.getMethodMetadata(REST_METHODS_WITH_PARAMETERS_KEY, constructor.prototype, op);
        }
        debug('  parameters for method %s: %j', op, params);
        if (params != null) {
            const bodyParams = params.filter(p => p && p.in === 'body');
            if (bodyParams.length > 1) {
                throw new Error('More than one body parameters found: ' + bodyParams);
            }
            params = context_1.DecoratorFactory.cloneDeep(params);
            operationSpec.parameters = params;
        }
        operationSpec['x-operation-name'] = op;
        if (!spec.paths[path]) {
            spec.paths[path] = {};
        }
        if (spec.paths[path][verb]) {
            // Operations from subclasses override those from the base
            debug(`  Overriding ${endpointName} - endpoint was already defined`);
        }
        debug(`  adding ${endpointName}`, operationSpec);
        spec.paths[path][verb] = operationSpec;
    }
    return spec;
}
/**
 * Get the controller spec for the given class
 * @param constructor Controller class
 */
function getControllerSpec(constructor) {
    let spec = context_1.MetadataInspector.getClassMetadata(REST_API_SPEC_KEY, constructor, { ownMetadataOnly: true });
    if (!spec) {
        spec = resolveControllerSpec(constructor);
        context_1.MetadataInspector.defineMetadata(REST_API_SPEC_KEY, spec, constructor);
    }
    return spec;
}
exports.getControllerSpec = getControllerSpec;
/**
 * Expose a Controller method as a REST API operation
 * mapped to `GET` request method.
 *
 * @param path The URL path of this operation, e.g. `/product/{id}`
 * @param spec The OpenAPI specification describing parameters and responses
 *   of this operation.
 */
function get(path, spec) {
    return operation('get', path, spec);
}
exports.get = get;
/**
 * Expose a Controller method as a REST API operation
 * mapped to `POST` request method.
 *
 * @param path The URL path of this operation, e.g. `/product/{id}`
 * @param spec The OpenAPI specification describing parameters and responses
 *   of this operation.
 */
function post(path, spec) {
    return operation('post', path, spec);
}
exports.post = post;
/**
 * Expose a Controller method as a REST API operation
 * mapped to `PUT` request method.
 *
 * @param path The URL path of this operation, e.g. `/product/{id}`
 * @param spec The OpenAPI specification describing parameters and responses
 *   of this operation.
 */
function put(path, spec) {
    return operation('put', path, spec);
}
exports.put = put;
/**
 * Expose a Controller method as a REST API operation
 * mapped to `PATCH` request method.
 *
 * @param path The URL path of this operation, e.g. `/product/{id}`
 * @param spec The OpenAPI specification describing parameters and responses
 *   of this operation.
 */
function patch(path, spec) {
    return operation('patch', path, spec);
}
exports.patch = patch;
/**
 * Expose a Controller method as a REST API operation
 * mapped to `DELETE` request method.
 *
 * @param path The URL path of this operation, e.g. `/product/{id}`
 * @param spec The OpenAPI specification describing parameters and responses
 *   of this operation.
 */
function del(path, spec) {
    return operation('delete', path, spec);
}
exports.del = del;
/**
 * Expose a Controller method as a REST API operation.
 *
 * @param verb HTTP verb, e.g. `GET` or `POST`.
 * @param path The URL path of this operation, e.g. `/product/{id}`
 * @param spec The OpenAPI specification describing parameters and responses
 *   of this operation.
 */
function operation(verb, path, spec) {
    return context_1.MethodDecoratorFactory.createDecorator(REST_METHODS_KEY, {
        verb,
        path,
        spec,
    });
}
exports.operation = operation;
const paramDecoratorStyle = Symbol('ParamDecoratorStyle');
/**
 * Check if the given type is `stream.Readable` or a subclasses of
 * `stream.Readable`
 * @param type JavaScript type function
 */
function isReadableStream(type) {
    if (typeof type !== 'function')
        return false;
    if (type === stream.Readable)
        return true;
    return isReadableStream(Object.getPrototypeOf(type));
}
/**
 * Get openapi type name for a JavaScript type
 * @param type JavaScript type
 */
function getTypeForNonBodyParam(type) {
    if (type === String) {
        return 'string';
    }
    else if (type === Number) {
        return 'number';
    }
    else if (type === Boolean) {
        return 'boolean';
    }
    else if (type === Array) {
        return 'array';
    }
    else if (isReadableStream(type)) {
        return 'file';
    }
    return 'string';
}
/**
 * Get openapi schema for a JavaScript type for a body parameter
 * @param type JavaScript type
 */
function getSchemaForBodyParam(type) {
    const schema = {};
    let typeName;
    if (type === String) {
        typeName = 'string';
    }
    else if (type === Number) {
        typeName = 'number';
    }
    else if (type === Boolean) {
        typeName = 'boolean';
    }
    else if (type === Array) {
        // item type cannot be inspected
        typeName = 'array';
    }
    else if (isReadableStream(type)) {
        typeName = 'file';
    }
    else if (type === Object) {
        typeName = 'object';
    }
    if (typeName) {
        schema.type = typeName;
    }
    else {
        schema.$ref = '#/definitions/' + type.name;
    }
    return schema;
}
/**
 * Describe an input parameter of a Controller method.
 *
 * `@param` can be applied to method itself or specific parameters. For example,
 * ```
 * class MyController {
 *   @get('/')
 *   @param(offsetSpec)
 *   @param(pageSizeSpec)
 *   list(offset?: number, pageSize?: number) {}
 * }
 * ```
 * or
 * ```
 * class MyController {
 *   @get('/')
 *   list(
 *     @param(offsetSpec) offset?: number,
 *     @param(pageSizeSpec) pageSize?: number,
 *   ) {}
 * }
 * ```
 * Please note mixed usage of `@param` at method/parameter level is not allowed.
 *
 * @param paramSpec Parameter specification.
 */
function param(paramSpec) {
    return function (target, member, descriptorOrIndex) {
        paramSpec = paramSpec || {};
        // Get the design time method parameter metadata
        const methodSig = context_1.MetadataInspector.getDesignTypeForMethod(target, member);
        const paramTypes = (methodSig && methodSig.parameterTypes) || [];
        const targetWithParamStyle = target;
        if (typeof descriptorOrIndex === 'number') {
            if (targetWithParamStyle[paramDecoratorStyle] === 'method') {
                // This should not happen as parameter decorators are applied before
                // the method decorator
                /* istanbul ignore next */
                throw new Error('Mixed usage of @param at method/parameter level' +
                    ' is not allowed.');
            }
            // Map design-time parameter type to the OpenAPI param type
            let paramType = paramTypes[descriptorOrIndex];
            if (paramType) {
                if (paramSpec.in !== 'body') {
                    if (!paramSpec.type) {
                        paramSpec.type = getTypeForNonBodyParam(paramType);
                    }
                }
                else {
                    paramSpec.schema = Object.assign(getSchemaForBodyParam(paramType), paramSpec.schema);
                }
            }
            if (paramSpec.type === 'array' ||
                (paramSpec.schema && paramSpec.schema.type === 'array')) {
                paramType = paramTypes[descriptorOrIndex];
                // The design-time type is `Object` for `any`
                if (paramType != null && paramType !== Object && paramType !== Array) {
                    throw new Error(`The parameter type is set to 'array' but the JavaScript type is ${paramType.name}`);
                }
            }
            targetWithParamStyle[paramDecoratorStyle] = 'parameter';
            context_1.ParameterDecoratorFactory.createDecorator(REST_PARAMETERS_KEY, paramSpec)(target, member, descriptorOrIndex);
        }
        else {
            if (targetWithParamStyle[paramDecoratorStyle] === 'parameter') {
                throw new Error('Mixed usage of @param at method/parameter level' +
                    ' is not allowed.');
            }
            targetWithParamStyle[paramDecoratorStyle] = 'method';
            RestMethodParameterDecoratorFactory.createDecorator(REST_METHODS_WITH_PARAMETERS_KEY, paramSpec)(target, member, descriptorOrIndex);
        }
    };
}
exports.param = param;
class RestMethodParameterDecoratorFactory extends context_1.MethodParameterDecoratorFactory {
}
(function (param) {
    param.query = {
        /**
         * Define a parameter of "string" type that's read from the query string.
         *
         * @param name Parameter name.
         */
        string: createParamShortcut('query', 'string'),
        /**
         * Define a parameter of "number" type that's read from the query string.
         *
         * @param name Parameter name.
         */
        number: createParamShortcut('query', 'number'),
        /**
         * Define a parameter of "integer" type that's read from the query string.
         *
         * @param name Parameter name.
         */
        integer: createParamShortcut('query', 'integer'),
        /**
         * Define a parameter of "boolean" type that's read from the query string.
         *
         * @param name Parameter name.
         */
        boolean: createParamShortcut('query', 'boolean'),
    };
    param.header = {
        /**
         * Define a parameter of "string" type that's read from a request header.
         *
         * @param name Parameter name, it must match the header name
         *   (e.g. `Content-Type`).
         */
        string: createParamShortcut('header', 'string'),
        /**
         * Define a parameter of "number" type that's read from a request header.
         *
         * @param name Parameter name, it must match the header name
         *   (e.g. `Content-Length`).
         */
        number: createParamShortcut('header', 'number'),
        /**
         * Define a parameter of "integer" type that's read from a request header.
         *
         * @param name Parameter name, it must match the header name
         *   (e.g. `Content-Length`).
         */
        integer: createParamShortcut('header', 'integer'),
        /**
         * Define a parameter of "boolean" type that's read from a request header.
         *
         * @param name Parameter name, it must match the header name,
         *   (e.g. `DNT` or `X-Do-Not-Track`).
         */
        boolean: createParamShortcut('header', 'boolean'),
    };
    param.path = {
        /**
         * Define a parameter of "string" type that's read from request path.
         *
         * @param name Parameter name matching one of the placeholders in the path
         *   string.
         */
        string: createParamShortcut('path', 'string'),
        /**
         * Define a parameter of "number" type that's read from request path.
         *
         * @param name Parameter name matching one of the placeholders in the path
         *   string.
         */
        number: createParamShortcut('path', 'number'),
        /**
         * Define a parameter of "integer" type that's read from request path.
         *
         * @param name Parameter name matching one of the placeholders in the path
         *   string.
         */
        integer: createParamShortcut('path', 'integer'),
        /**
         * Define a parameter of "boolean" type that's read from request path.
         *
         * @param name Parameter name matching one of the placeholders in the path
         *   string.
         */
        boolean: createParamShortcut('path', 'boolean'),
    };
    param.formData = {
        /**
         * Define a parameter of "string" type that's read
         * from a field in the request body.
         *
         * @param name Parameter name.
         */
        string: createParamShortcut('formData', 'string'),
        /**
         * Define a parameter of "number" type that's read
         * from a field in the request body.
         *
         * @param name Parameter name.
         */
        number: createParamShortcut('formData', 'number'),
        /**
         * Define a parameter of "integer" type that's read
         * from a field in the request body.
         *
         * @param name Parameter name.
         */
        integer: createParamShortcut('formData', 'integer'),
        /**
         * Define a parameter of "boolean" type that's read
         * from a field in the request body.
         *
         * @param name Parameter name.
         */
        boolean: createParamShortcut('formData', 'boolean'),
    };
    /**
     * Define a parameter that's set to the full request body.
     *
     * @param name Parameter name
     * @param schema The schema defining the type used for the body parameter.
     */
    param.body = function (name, schema) {
        return param({ name, in: 'body', schema });
    };
    /**
     * Define a parameter of `array` type
     *
     * @example
     * ```ts
     * export class MyController {
     *   @get('/greet')
     *   greet(@param.array('names', 'query', 'string') names: string[]): string {
     *     return `Hello, ${names}`;
     *   }
     * }
     * ```
     * @param name Parameter name
     * @param source Source of the parameter value
     * @param itemSpec Item type for the array or the full item object
     */
    param.array = function (name, source, itemSpec) {
        const items = typeof itemSpec === 'string' ? { type: itemSpec } : itemSpec;
        if (source !== 'body') {
            return param({ name, in: source, type: 'array', items });
        }
        else {
            return param({ name, in: source, schema: { type: 'array', items } });
        }
    };
})(param = exports.param || (exports.param = {}));
function createParamShortcut(source, type) {
    // TODO(bajtos) @param.IN.TYPE('foo', {required: true})
    return (name) => {
        return param({ name, in: source, type });
    };
}
//# sourceMappingURL=controller-spec.js.map